/*-------------------- Collective  Communications Routines --------------*
 File : collective.c
   Date : 2/16/95

--------------------------------------------------------------------*/
#include "scarab.h"




/****************************************************************************/
void preprocessing_comm(
int sendcount[],
int recvcount[],
int sdispls[],
int rdispls[],
int *total_send,
int *total_recv)
{

  int i,tot_s,tot_r;


  AlltoAll(sendcount,1,sizeof(int),recvcount,1);  /*  Using LP */

  /*-- Calculating the sdispls and rdispls and the send buffer  --*/

  sdispls[0] = rdispls[0] = 0;
  tot_r = recvcount[0];
  tot_s = sendcount[0];
  for (i=1 ; i < PS ; i++)  {
    sdispls[i] = sdispls[i-1] + sendcount[i-1];
    rdispls[i] = rdispls[i-1] + recvcount[i-1];
    tot_r += recvcount[i];
    tot_s += sendcount[i];
  }
 
  *total_recv = tot_r;
  *total_send = tot_s;
}

/*-- All-to-all with uniform message sizes --*/
int AlltoAll(sendbuf,count,datatypesize,recvbuf, algorithm)
  void *sendbuf;    /* address of the send buffer */
  int   count;   /* Number of elements send/receive by each processor */
  int   datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;   /* address of the receive buffer */
  int   algorithm;  /* the algorithm number */
 { int i,ind,total;
   char *ptrsend,*ptrrecv;


  if (ArgTypeCheck !=0 )  {
    /* Check for error and report it,if any, */
  }
  if ( algorithm == 0 )  {
    /* Ask the expert system to determine the "best" algorithm */
    algorithm = Expert_System();
  }
  /* Perform the local copying */
  ind = Self*count*datatypesize;
   /* Problem : pointer of type void, how to do the copying */
  ptrsend = sendbuf;
  ptrrecv = recvbuf;
  ptrsend += ind;
  ptrrecv += ind;
  total = count*datatypesize;
  memcpy(ptrrecv,ptrsend,total);
  /* Preprocessing for a certain algorithm if needed */
  switch (algorithm) {
  case 1 : Error = LP_All(sendbuf,count,datatypesize,recvbuf);
           break;
  case 2 : Error = AC_All(sendbuf,count,datatypesize,recvbuf);
           break;
  case 3 : Error = DS_All(sendbuf,count,datatypesize,recvbuf);
           break;
  default : Error = -1;
            break;
  }
  return(Error);
 }


/*--  LP_Xor : Implementation of All-to-All ----*/
int LP_All(sendbuf,count,datatypesize,recvbuf)
  void *sendbuf;    /* address of the send buffer */
  int   count;   /* Number of elements send/receive by each processor */
  int   datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;  /* address of the receive buffer */
{
  int i,partner,ind,des,total;
  char *ptrsend,*ptrrecv;
  

  ind =  0;
  ptrsend = sendbuf;
  ptrrecv = recvbuf;
  total = count*datatypesize;
  for (i=1; i < PS ; i++)  {
   partner = Self ^ i;  /* Find the partner by Xoring */
   des = partner * count*datatypesize;
#if MACHINE == nCube
   Status = GenAsySend(&(ptrsend[des]),total,partner,ANY_TAG);
   Error = GenSynReceive(&(ptrrecv[des]),total,partner,ANY_TAG);
#else 
   Status= GenAsyReceive(&(ptrrecv[des]),total,partner,ANY_TAG);
   Error  = GenSynSend(&(ptrsend[des]),total,partner,ANY_TAG);

#endif
   GenAsyHanlde(Status);
/*   GenBarrier(); */

  }
  return(Error);
}


/*--  Async. Comm. : Implementation of All-to-All ----*/
int AC_All(sendbuf,count,datatypesize,recvbuf)
  void *sendbuf;    /* address of the send buffer */
  int   count;   /* Number of elements send/receive by each processor */
  int   datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;  /* address of the receive buffer */
{ int i,dest;
  long *rmsgid; /* It should be PS,but Paragon did not accept it */
  int total;
  char *ptrsend,*ptrrecv;


  total = count*datatypesize;
                /* Post all receiving messages */
  ptrsend = sendbuf;
  ptrrecv = recvbuf;
  rmsgid = (long *) malloc((unsigned)(sizeof(long)*PS));
  for (i=0; i < Self ; i++)
#if MACHINE == nCube
   rmsgid[i] = GenAsySend(&(ptrsend[total*i]),total,i,Self);   
#else
   rmsgid[i] = GenAsyReceive(&(ptrrecv[total*i]),total,i,i);
#endif
  for (i=Self+1; i < PS ; i++)
#if MACHINE == nCube
   rmsgid[i] = GenAsySend(&(ptrsend[total*i]),total,i,Self);   
#else
   rmsgid[i] = GenAsyReceive(&(ptrrecv[total*i]),total,i,i);
#endif


                 /* Send all message */
  srand48(Self*100+2);
  dest = lrand48() % PS;
  for (i=1; i < PS ; i++) {
   if (dest == Self ) dest = ( dest + 1 ) % PS;
#if MACHINE == nCube
   GenSynReceive(&(ptrrecv[total*dest]),total,dest,dest);
#else
   GenSynSend(&(ptrsend[total*dest]),total,dest,Self);
#endif
   dest = ( dest + 1) % PS;
  }
                /* Wait for receiving the messages */
  for (i=0; i < Self ; i++)
   GenAsyHanlde(rmsgid[i]);
  for (i=Self+1; i < PS ; i++)
   GenAsyHanlde(rmsgid[i]);
  free((char *) rmsgid);
  GenBarrier();

}



/*--  DS : Implementation of All-to-All ----*/

             /*-----  Global Variable for DS   ----*/
int freeport[2];  /* to keep track of free ports */
int busy;         /* 1 -> I am busy. 0 -> I am free */
int busy_dest;    /* 0 -> Use rport 0
                     1 -> Use rport 1
                     2 -> the destination is busy */
int get_reply;    /* 0 -> I am waiting for reply.
                     1 -> I have received the reply */
#define Maximum 32  /* Based on it, the processor determine if it can use
                      2 ports at the same time */
             /*-----  Handlers     -----*/
 /* This handler is called when rport counter reacHes zero */
void DS_done(rport)
 int rport;
{
  busy--;
  freeport[--rport] = 0;  /* This port is free */
}
/*  This handler is called upon receiving a reply active message */
void DS_reply_All(rport)
int rport;
{
 get_reply = 1;  /* I got a reply */
 busy_dest = rport;
}
/*   This handler is called upon receiving a request active message */
void DS_request_All(source,recvbuf,no_bytes)
int source;
void *recvbuf;
int no_bytes;
{
 int rport,ok;

  ok = 0;
  if ( busy == 0 ) {
    ok = 1;       /* I can receive the message */
    rport = 0;    /* use rport number 0 */
  }
  else           /* check if i can use 2 rports at the same time */
   if ( busy == 1 ) {  /* I am using only 1 rport now */
     if (freeport[0] == 1 )
      rport = 0;    /* the busy port */
     else rport = 1;
     if ( Get_rport_byte_counter(rport+1) <=  Maximum ) {
        /* I can use 2 rports */
        ok = 1;
        rport = (rport + 1) % 2;  /* the free rport */
      }    
    }

 if ( ok == 1 ) {  /* I can receive the message */
  freeport[rport++] = 1;  /* this rport is busy */
  busy++;
  /* Problem : Checking the alignment */
     /*----  Setting the rport for receving -----*/
  Set_rport_address(rport,recvbuf);
  Set_rport_handler(rport,DS_done);
  Set_rport_byte_counter(rport,no_bytes);
     /*--- Sending a positive reply -----*/
  Gen_reply(source,DS_reply_All,rport);
  }
  else   /*----- Sending negative reply  -----*/
   Gen_reply(source,DS_reply_All,3);
}

int DS_All(sendbuf,count,datatypesize,recvbuf)
  void *sendbuf;    /* address of the send buffer */
  int   count;   /* Number of elements send/receive by each processor */
  int   datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;  /* address of the receive buffer */
{
 int i,dest,*dest_arr; /* It should be PS-1, but Paragon did
    not accept it */
 int next,rem_msg;
 char *ptrsend,*ptrrecv;

  Set_Global_Or(1);    /* Start sending process */
  srand48(Self*100+2);
  freeport[0] = freeport[1] = 0;  /* both rports are free */
  busy = 0;   /* I am free */
  dest = 0;
  if ( Allocate_rport(1)==-1) printf("error...1\n");
  if ( Allocate_rport(2)==-1) printf("Error...2\n");
  dest_arr = (int *) malloc((unsigned)(sizeof(int)*(PS-1)));
  GenBarrier();  /* Debuging */
  rem_msg = PS-1;
  for (i= 0; i < rem_msg ; i++)  {
     if ( dest == Self ) dest = (dest + 1) % PS;
     dest_arr[i] = dest;
     dest = ( dest + 1 ) % PS;
  }
  next = lrand48() % PS;
  while ( rem_msg > 0 )  {
    next = ( next + 1 ) % rem_msg;
    dest = dest_arr[next];
       /*-- send a request to the current destination --*/
    get_reply = 0;
    ptrrecv = recvbuf;
    ptrrecv += count*Self*datatypesize;
    Gen_request(dest,DS_request_All,Self,ptrrecv,
                count*datatypesize);
       /*-- Wait for a reply ---*/
    Gen_Poll_until(&get_reply,1);
    if  (busy_dest !=3 ) {   /*  start sending the message */
     ptrsend = sendbuf;
     ptrsend += count*dest*datatypesize;
     Gen_send_rport(dest,busy_dest,0,ptrsend,ptrrecv,count,datatypesize);
     dest_arr[next] = dest_arr[--rem_msg];
     }
   }  /* While loop */
  Set_Global_Or(0);   /* I have done sending all the messages */
  while ( ( busy != 0 )  || ( Get_Global_Or() == 1 ) )
    Gen_Poll();
  /*  Free the rports */
  free((char *) dest_arr);
  Free_rport(1);
  Free_rport(2);
  GenBarrier();
}

   /*=================================================*/
/*-- Many-to-Many with un-uniform message sizes and contigous buffer--*/
int ManytoMany_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,recvcount,
                 rdispls,algorithm,algorithm1,algorithm2)
  void *sendbuf;    /* address of the send buffer */
  int  *sendcount;   /* Number of elements send to each processor, 
                        This is an array of size PS */
  int  *sdispls;   /* the displacement of each send block from the beginning of
               the sendbuf*/
  int  datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;   /* address of the receive buffer */
  int  *recvcount;   /* No. of elements receive from each processor. 
                        This is an array of size PS */
  int  *rdispls;   /* the displacement of each recv block from the beginning 
                      of the recvbuf */
  int  algorithm;   /* the main algorithm number */
  int  algorithm1;   /* the first stage algorithm, in case of two stage algorithm*/
  int  algorithm2 ; /* the second stage algorithm, in case of two stage algorithm */

 { int i,inds,indr,total;
   char *ptrsend,*ptrrecv;


  if (ArgTypeCheck !=0 )  {
    /* Check for error and report it,if any, */
  }
  if ( algorithm == 0 )  {
    /* Ask the expert system to determine the "best" algorithm */
    algorithm = Expert_System();
  }
  /* Perform the local copying */
  inds = sdispls[Self];
  indr = rdispls[Self]; 
   /* Problem : pointer of type void, how to do the copying */
  ptrsend = sendbuf;
  ptrrecv = recvbuf;

  total = sendcount[Self]*datatypesize; 
  ptrsend += sdispls[Self]*datatypesize;
  ptrrecv += rdispls[Self]*datatypesize;
  memcpy(ptrrecv,ptrsend,total); 
 
  /* Preprocessing for a certain algorithm if needed */
  switch (algorithm) {
  case 1 : Error = LP_Many_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,
                               recvcount,rdispls);

           break;
  case 2 : Error = AC_Many_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,
                               recvcount,rdispls);
           break;
  case 3 : Error = DS_Many_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,
                               recvcount,rdispls);
           break;
  case 4 : Error = Two_Stage_Many_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,
                   recvcount,rdispls,algorithm1,algorithm2);
           break;
  default : Error = -1;
            break;
  }
  return(Error);
 }


/*--  LP_Xor : Implementation of Many-to-Many with contigous buffer ----*/
int LP_Many_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,recvcount,
                     rdispls)
  void *sendbuf;    /* address of the send buffer */
  int  *sendcount;   /* Number of elements send to each processor, 
                        This is an array of size PS */
  int  *sdispls;   /* the displacement of each send block from the beginning of
               the sendbuf*/
  int  datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;   /* address of the receive buffer */
  int  *recvcount;   /* No. of elements receive from each processor. 
                        This is an array of size PS */
  int  *rdispls;   /* the displacement of each recv block from the beginning 
                      of the recvbuf */
{
  int i,partner,ind,des,total;
  char *ptrsend,*ptrrecv;
  

  ptrsend = sendbuf;
  ptrrecv = recvbuf;

  for (i=1; i < PS ; i++)  {
   partner = Self ^ i;  /* Find the partner by Xoring */
#if MACHINE == nCube
   if ( sendcount[partner] > 0 )  {
     des = sdispls[partner]*datatypesize; 
     Status = GenAsySend((ptrsend+des),sendcount[partner]*datatypesize,
                           partner,Self);
   }
   if ( recvcount[partner] > 0 ) {
      des = rdispls[partner]*datatypesize;
      Error = GenSynReceive((ptrrecv+des),recvcount[partner]*datatypesize,
              partner,partner);
   }
   if ( sendcount[partner] > 0 )
          GenAsyHanlde(Status);
#else
   if ( recvcount[partner] > 0 ) {
      des = rdispls[partner]*datatypesize;
      Status = GenAsyReceive((ptrrecv+des),recvcount[partner]*datatypesize,
              partner,partner);
   }
   if ( sendcount[partner] > 0 )  {
     des = sdispls[partner]*datatypesize; 
     Error = GenSynSend((ptrsend+des),sendcount[partner]*datatypesize,
                           partner,Self);
   }
   if ( recvcount[partner] > 0 )
          GenAsyHanlde(Status);

#endif
#if MACHINE==CM_5
   GenBarrier();
#endif

  }
#if MACHINE==Paragon
   GenBarrier();
#endif
  return(Error);
}


/*--  Async. Comm. : Implementation of Many-to-Many with contigous buffer----*/
int AC_Many_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,recvcount,
                     rdispls)
  void *sendbuf;    /* address of the send buffer */
  int  *sendcount;   /* Number of elements send to each processor, 
                        This is an array of size PS */
  int  *sdispls;   /* the displacement of each send block from the beginning of
               the sendbuf*/
  int  datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;   /* address of the receive buffer */
  int  *recvcount;   /* No. of elements receive from each processor. 
                        This is an array of size PS */
  int  *rdispls;   /* the displacement of each recv block from the beginning 
                      of the recvbuf */

{ int i,dest;
  long *rmsgid; /* It should be PS,but Paragon did not accept it */
  int total;
  char *ptrsend,*ptrrecv;


                /* Post all receiving messages */
  ptrsend = sendbuf;
  ptrrecv = recvbuf;
  rmsgid = (long *) malloc((unsigned)(sizeof(long)*PS));
#if MACHINE == nCube
  for (i=0; i < Self ; i++)
   if ( sendcount[i] > 0 ) 
     rmsgid[i] = GenAsySend(&ptrsend[sdispls[i]*datatypesize],
                         sendcount[i]*datatypesize,i,Self*1000+i);

  for (i=Self+1; i < PS ; i++)
   if ( sendcount[i] > 0 ) 
     rmsgid[i] = GenAsySend(&ptrsend[sdispls[i]*datatypesize],
                         sendcount[i]*datatypesize,i,Self*1000+i);

                 /* Send all message */
  srand48(Self*100+2);
  dest = lrand48() % PS;
  for (i=1; i < PS ; i++) {
   if (dest == Self ) dest = ( dest + 1 ) % PS;
   if ( recvcount[dest] > 0 )
     GenSynReceive(&ptrrecv[rdispls[dest]*datatypesize],recvcount[dest]*
                               datatypesize,dest,dest*1000+Self);
   dest = ( dest + 1) % PS;
  }
                /* Wait for receiving the messages */
  for (i=0; i < Self ; i++)
   if ( sendcount[i] > 0 )    GenAsyHanlde(rmsgid[i]);
  for (i=Self+1; i < PS ; i++)
   if ( sendcount[i] > 0 )    GenAsyHanlde(rmsgid[i]);
#else

  for (i=0; i < Self ; i++)
   if ( recvcount[i] > 0 ) 
     rmsgid[i] = GenAsyReceive(&ptrrecv[rdispls[i]*datatypesize],recvcount[i]*
                               datatypesize,i,i*1000+Self);
  for (i=Self+1; i < PS ; i++)
   if ( recvcount[i] > 0 ) 
     rmsgid[i] = GenAsyReceive(&ptrrecv[rdispls[i]*datatypesize],recvcount[i]*
                               datatypesize,i,i*1000+Self);
                 /* Send all message */
  srand48(Self*100+2);
  dest = lrand48() % PS;
  for (i=1; i < PS ; i++) {
   if (dest == Self ) dest = ( dest + 1 ) % PS;
   if ( sendcount[dest] > 0 )
     GenSynSend(&ptrsend[sdispls[dest]*datatypesize],sendcount[dest]*
                datatypesize,dest,Self*1000+dest);
   dest = ( dest + 1) % PS;
  }
                /* Wait for receiving the messages */
  for (i=0; i < Self ; i++)
   if ( recvcount[i] > 0 )    GenAsyHanlde(rmsgid[i]);
  for (i=Self+1; i < PS ; i++)
   if ( recvcount[i] > 0 )    GenAsyHanlde(rmsgid[i]);
#endif
  free((char *) rmsgid);

  GenBarrier();

}

    /*=========   DS Handlers and global varaibles   =========*/

/*    Do not forget to free the ports after using them */
int  *Rdispls;
char *Recvbuf;
char *DestRecvBuf;
int  Datatypesize;
int *Rcount;
/*  This handler is called upon receiving a reply active message */
void DS_reply_Many_Con(rport,ptrrecv)
int rport;
char *ptrrecv;  
{
 get_reply = 1;  /* I got a reply */
 busy_dest = rport;
 DestRecvBuf = ptrrecv;
}
/*   This handler is called upon receiving a request active message */
void DS_request_Many_Con(source,no_bytes)
int source;

int no_bytes;
{
 int rport,ok;

  ok = 0;
  if ( busy == 0 ) {
    ok = 1;       /* I can receive the message */
    rport = 0;    /* use rport number 0 */
  }
  else           /* check if i can use 2 rports at the same time */
   if ( busy == 1 ) {  /* I am using only 1 rport now */
     if (freeport[0] == 1 )
      rport = 0;    /* the busy port */
     else rport = 1;
     if ( Get_rport_byte_counter(rport+1) <=  Maximum ) {
        /* I can use 2 rports */
        ok = 1;
        rport = (rport + 1) % 2;  /* the free rport */
      }    
    }

 if ( ok == 1 ) {  /* I can receive the message */
  freeport[rport++] = 1;  /* this rport is busy */
  busy++;

  /* Problem : Checking the alignment */
     /*----  Setting the rport for receving -----*/
  Set_rport_address(rport,&Recvbuf[Rdispls[source]*Datatypesize]);
  Set_rport_handler(rport,DS_done);
   /*  Here we may check the number of bytes; no_bytes ?= rcount[source]*Data*/
  Set_rport_byte_counter(rport,no_bytes);
     /*--- Sending a positive reply -----*/
  Gen_reply(source,DS_reply_Many_Con,rport,&Recvbuf[Rdispls[source]*Datatypesize]);
  }
  else   /*----- Sending negative reply  -----*/
   Gen_reply(source,DS_reply_Many_Con,3);
}

/*--  DS : Implementation of Many-to-Many with contigous buffer ----*/
int DS_Many_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,recvcount,
                     rdispls)
  void *sendbuf;    /* address of the send buffer */
  int  *sendcount;   /* Number of elements send to each processor,
                        This is an array of size PS */
  int  *sdispls;   /* the displacement of each send block from the beginning of
               the sendbuf*/
  int  datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;   /* address of the receive buffer */
  int  *recvcount;   /* No. of elements receive from each processor.
                        This is an array of size PS */
  int  *rdispls;   /* the displacement of each recv block from the beginning
                      of the recvbuf */
{
 int i,dest,*dest_arr; /* It should be PS-1, but Paragon did
    not accept it */
 int next,rem_msg;
 char *ptrsend,*ptrrecv;

  Set_Global_Or(1);    /* Start sending process */
  srand48(Self*100+2);
  freeport[0] = freeport[1] = 0;  /* both rports are free */
  busy = 0;   /* I am free */
  dest = 0;
  Rdispls = rdispls;
  Datatypesize = datatypesize;
  Recvbuf = recvbuf;
  Rcount = recvcount;
  dest_arr = (int *) malloc((unsigned)(sizeof(int)*(PS-1)));
  if ( Allocate_rport(1)==-1) printf("error...1\n");
  if ( Allocate_rport(2)==-1) printf("Error...2\n");

  GenBarrier();  /* For Debuging */
  rem_msg = 0;
  for (i= 0; i < Self ; i++)  
       if ( sendcount[i] > 0 ) dest_arr[rem_msg++] = i;
  for (i= Self+1; i < PS ; i++)  
       if ( sendcount[i] > 0 ) dest_arr[rem_msg++] = i;
  next = lrand48() % PS;
  while ( rem_msg > 0 )  {
    next = ( next + 1 ) % rem_msg;
    dest = dest_arr[next];
       /*-- send a request to the current destination --*/
    get_reply = 0;
    Gen_request(dest,DS_request_Many_Con,Self,sendcount[dest]*datatypesize);
       /*-- Wait for a reply ---*/
    Gen_Poll_until(&get_reply,1);
    if  (busy_dest !=3 ) {   /*  start sending the message */
     ptrsend = sendbuf;
     ptrsend += sdispls[dest]*datatypesize;

     Gen_send_rport(dest,busy_dest,0,ptrsend,DestRecvBuf,sendcount[dest],datatypesize);
     dest_arr[next] = dest_arr[--rem_msg];
     }
   }  /* While loop */
  Set_Global_Or(0);   /* I have done sending all the messages */
  while ( ( busy != 0 )  || ( Get_Global_Or() == 1 ) )
    Gen_Poll();
/* Free rports */
  if ( Free_rport(1)==-1) printf(" Err....");
  if ( Free_rport(2)==-1) printf(" Err2...");
  free((char *) dest_arr);
  GenBarrier();
}

/*-- Two Stage Alg. : Implementation of Many-to-Many with contigous buffer ----*/
                  /*--  Handlers and variables --*/
int *recv_msg_len;
int *send1_msg_len;
int *send2_msg_len;
typedef char *charptr;

charptr *send1_msg_ptr;
charptr *send2_msg_ptr;
char *recv_stage1;
char *recv_stage2;
int  length;
int *Recv_msg_len;
int *Destrecv_msg_len;


/*-----------   LP_Xor for Two stage algorithm Without Reshuffling   ------*/
int Two_stg_LP_WOR(sendbuf,recvbuf,send_msg_len,send_msg_ptr,
                    recv_msg_len,datatypesize)
void *sendbuf;
void *recvbuf;
int  *send_msg_len;
charptr *send_msg_ptr;
int  *recv_msg_len;
int datatypesize;
{ int i,j,partner,total,ind;
  char *ptrrecv;  
  char *ptrsend;
  long *Ids;
  long Id;
  int  PS_plus;

  PS_plus = PS +1;
  ptrsend = sendbuf;
  ptrrecv = recvbuf;
  Ids = (long *) malloc((unsigned)(sizeof(long)*PS));

  total = sizeof(int)*(PS+1);
  for (j=1; j < PS ; j++)  {
   partner = Self ^ j;  /* Xoring */
   Id = GenAsySend(&(send_msg_len[partner*PS_plus]),total,partner,PS_plus);
   Error= GenSynReceive(&(recv_msg_len[partner*PS_plus]),total,partner,PS_plus);
   GenAsyHanlde(Id);

   /*sending the sub messages*/
   if (send_msg_len[partner*PS_plus] >0)  {
    for (i=0; i<PS;i++)
     if (send_msg_len[partner*PS_plus+(i+1)] >0)


      Ids[i] = GenAsySend((void *) send_msg_ptr[partner*PS+i],
                     send_msg_len[partner*PS_plus+(i+1)]*datatypesize,partner,i);
    }

   if ( recv_msg_len[partner*PS_plus] > 0 ) {
    ind = 0;
    for (i=0 ; i < PS ; i++)
      if ( recv_msg_len[partner*PS_plus+(i+1)] > 0 ) {
        Error = GenSynReceive( (ptrrecv+(datatypesize*(length*partner+ind))),
                  recv_msg_len[partner*PS_plus+(i+1)]*datatypesize,partner,i); 

        ind += recv_msg_len[partner*PS_plus+(i+1)];
      }
    }
   if (send_msg_len[partner*PS_plus] >0)  {
    for (i=0; i<PS;i++)
     if (send_msg_len[partner*PS_plus+(i+1)] >0)
        GenAsyHanlde(Ids[i]);
    }
    GenBarrier();
   }
  free((char *) Ids);
  return(Error);
}

/*-----------   LP_Xor for Two stage algorithm With Reshuffling   ------*/
int Two_stg_LP_WR(sendbuf,recvbuf,send_msg_len,send_msg_ptr,
                    recv_msg_len,datatypesize)
void *sendbuf;
void *recvbuf;
int  *send_msg_len;
charptr *send_msg_ptr;
int  *recv_msg_len;
int datatypesize;
{ int i,j,k,partner,total,ind;
  char *ptrrecv;  
  char *ptrsend;
  long Id;
  char *temp;
  int  PS_plus;

  PS_plus = PS +1;

  ptrrecv = recvbuf;
  ptrsend = (char *) malloc((unsigned)(datatypesize*length));

  total = sizeof(int)*(PS+1);
  for (j=1; j < PS ; j++)  {
   partner = Self ^ j;  /* Xoring */
   Id = GenAsySend(&(send_msg_len[partner*PS_plus]),total,partner,PS_plus);
   Error= GenSynReceive(&(recv_msg_len[partner*PS_plus]),total,partner,PS_plus);
   GenAsyHanlde(Id);

   /*sending the message*/
   if (send_msg_len[partner*PS_plus] >0)  {
       /* Reshuffling */
    ind = 0;
    temp = ptrsend;
    for (i=0; i<PS;i++) 
     if ( send_msg_len[partner*PS_plus+(i+1)] > 0 ) {
     memcpy(temp,send_msg_ptr[partner*PS+i],send_msg_len[partner*PS_plus+(i+1)]*datatypesize);
/*     for (k=0;  k < send_msg_len[partner*PS_plus+(i+1)]*datatypesize ; k++)
       temp[ind+k] =  *(send_msg_ptr[partner*PS+i]+k); */
     temp +=  send_msg_len[partner*PS_plus+(i+1)]*datatypesize;
   }

     Id = GenAsySend( ptrsend,
                     send_msg_len[partner*PS_plus]*datatypesize,partner,ANY_TAG);
  }
   if ( recv_msg_len[partner*PS_plus] > 0 )
        Error = GenSynReceive( (ptrrecv+(datatypesize*(length*partner))),
                  recv_msg_len[partner*PS_plus]*datatypesize,partner,ANY_TAG);

   if (send_msg_len[partner*PS_plus] >0)     GenAsyHanlde(Id);
   GenBarrier();
  }
  free((char *) ptrsend);
  return(Error);
}

/*-----------   DS for Two stage algorithm   --------------*/
  /* -- Handlers --*/
void Two_Stg_Done(rport)
int rport;
{
  int ind;


    if ( rport > 2 ) ind = rport - 3 ;
    else ind = rport - 1;
    freeport[ind]--;
    if ( freeport[ind] == 0 )  busy--;
}

void Two_Stg_Reply(rport,ptrrecv,ptrrecvlen)
int rport;
char *ptrrecv;
int *ptrrecvlen;
{
  get_reply = 1;
  busy_dest = rport;
  DestRecvBuf = ptrrecv;
  Destrecv_msg_len = ptrrecvlen;
}

/*   This handler is called upon receiving a request active message */
void Two_Stg_Request(source,no_bytes)
int source;
int no_bytes;

{
 int rport,ok;
 char *ptrrecv;
 int  *ptrrecvlen;

  ok = 0;
  if ( busy == 0 ) {
    ok = 1;       /* I can receive the message */
    rport = 0;    /* use rport number 0 */
  }
  else           /* check if i can use 2 rports at the same time */
   if ( busy == 1 ) {  /* I am using only 1 rport now */
     if (freeport[0] != 0 )
      rport = 0;    /* the busy port */
     else rport = 1;
     if ( Get_rport_byte_counter(rport+1) <=  Maximum ) {
        /* I can use 2 rports */
        ok = 1;
        rport = (rport + 1) % 2;  /* the free rport */
      }
    }

 if ( ok == 1 ) {  /* I can receive the message */
  freeport[rport++] = 2;  /*  rport & rport+2 are busy */
  busy++;
     /*----  Setting the rports for receving -----*/
  ptrrecv = Recvbuf + source*length*Datatypesize;
  ptrrecvlen = &Recv_msg_len[source*(PS+1)];

  Set_rport_address(rport,ptrrecv);
  Set_rport_handler(rport,Two_Stg_Done);
   /*  Here we may check the number of bytes; no_bytes ?= rcount[source]*Data*/
  Set_rport_byte_counter(rport,no_bytes);
  Set_rport_address(rport+2,ptrrecvlen);
  Set_rport_handler(rport+2,Two_Stg_Done);
  Set_rport_byte_counter(rport+2,(PS+1)*sizeof(int));
     /*--- Sending a positive reply -----*/
  Gen_reply(source,Two_Stg_Reply,rport,ptrrecv,ptrrecvlen);
  }
  else   /*----- Sending negative reply  -----*/
   Gen_reply(source,Two_Stg_Reply,3);
}

         /*=====    DS wothout Reshuffling ===============*/
int Two_stg_DS_WOR(sendbuf,recvbuf,send_msg_len,send_msg_ptr,
                    recv_msg_len,length,datatypesize)
void *sendbuf;
void *recvbuf;
int  *send_msg_len;
charptr  *send_msg_ptr;
int  *recv_msg_len;
int  length;
int datatypesize;
{
 int i,j,dest,*dest_arr; 

 int next,rem_msg;
 char *ptrsend,*ptrrecv;
 int offset,PS_plus;

  Set_Global_Or(1);    /* Start sending process */
  srand48(Self*100+2);
  freeport[0] = freeport[1] = 0;  /* both rports are free */
  busy = 0;   /* I am free */
  PS_plus = PS +1;
  if ( Allocate_rport(1)==-1) printf("error...1\n");
  if ( Allocate_rport(2)==-1) printf("Error...2\n");
  if ( Allocate_rport(3)==-1) printf("error...1\n");
  if ( Allocate_rport(4)==-1) printf("Error...2\n");
  dest_arr = (int *) malloc((unsigned)(sizeof(int)*(PS-1)));

  rem_msg = 0;
  for (i=0 ; i < PS ; i++)
    for (j=0 ; j <= PS ; j++)
       recv_msg_len[i*PS_plus+j] = 0;
  for (i= 0; i < Self ; i++)  {
       if ( send_msg_len[i*PS_plus] > 0 ) dest_arr[rem_msg++] = i;
      /* recv_msg_len[i][0] = 0; */
  }
  for (i= Self+1; i < PS ; i++)  {
       if ( send_msg_len[i*PS_plus] > 0 ) dest_arr[rem_msg++] = i;
     /*  recv_msg_len[i][0] = 0;*/
  }
/*  recv_msg_len[Self][0] =0; */
  Datatypesize = datatypesize;
  Recvbuf = recvbuf;
  Recv_msg_len = recv_msg_len;
  next = lrand48() % PS;
  while ( rem_msg > 0 )  {
    next = ( next + 1 ) % rem_msg;
    dest = dest_arr[next];
       /*-- send a request to the current destination --*/
    get_reply = 0;
    Gen_request(dest,Two_Stg_Request,Self,send_msg_len[dest*PS_plus]*datatypesize);
       /*-- Wait for a reply ---*/
    Gen_Poll_until(&get_reply,1);
    if  (busy_dest !=3 ) {   /*  start sending the message */

         /* sending the submessage lengths */
     Gen_send_rport(dest,busy_dest+2,0,&send_msg_len[dest*PS_plus],
                   Destrecv_msg_len,PS+1,sizeof(int));
     offset = 0;
     for (i=0 ; i < PS ; i++)
        /* Sending the submessages one at a time */
        if ( send_msg_len[dest*PS_plus+(i+1)] > 0 )  {

          Gen_send_rport(dest,busy_dest,offset,send_msg_ptr[dest*PS+i],
                         DestRecvBuf,send_msg_len[dest*PS_plus+(i+1)],datatypesize);
          offset += send_msg_len[dest*PS_plus+(i+1)];
        }

     dest_arr[next] = dest_arr[--rem_msg];
     }
   }  /* While loop */
  Set_Global_Or(0);   /* I have done sending all the messages */
  while ( ( busy != 0 )  || ( Get_Global_Or() == 1 ) )
    Gen_Poll();
/* Free rports */
  if ( Free_rport(1)==-1) printf(" Err....");
  if ( Free_rport(2)==-1) printf(" Err2...");
  if ( Free_rport(3)==-1) printf(" Err....");
  if ( Free_rport(4)==-1) printf(" Err2...");
  free((char *) dest_arr);
  GenBarrier();
}

         /*=====    DS with Reshuffling ===============*/
int Two_stg_DS_WR(sendbuf,recvbuf,send_msg_len,send_msg_ptr,
                    recv_msg_len,length,datatypesize)
void *sendbuf;
void *recvbuf;
int  *send_msg_len;
charptr  *send_msg_ptr;
int  *recv_msg_len;
int  length;
int datatypesize;
{
 int i,j,k,dest,*dest_arr; 

 int next,rem_msg;
 char *ptrsend,*ptrrecv;
 int ind,PS_plus;
 char *temp;

  Set_Global_Or(1);    /* Start sending process */
  srand48(Self*100+2);
  freeport[0] = freeport[1] = 0;  /* both rports are free */
  busy = 0;   /* I am free */
  PS_plus = PS +1;
  if ( Allocate_rport(1)==-1) printf("error...1\n");
  if ( Allocate_rport(2)==-1) printf("Error...2\n");
  if ( Allocate_rport(3)==-1) printf("error...1\n");
  if ( Allocate_rport(4)==-1) printf("Error...2\n");
  dest_arr = (int *) malloc((unsigned)(sizeof(int)*(PS-1)));
  ptrsend = (char *) malloc((unsigned)(datatypesize*length));
  rem_msg = 0;
  for (i=0 ; i < PS ; i++)
    for (j=0 ; j <= PS ; j++)
       recv_msg_len[i*PS_plus+j] = 0;
  for (i= 0; i < Self ; i++)  {
       if ( send_msg_len[i*PS_plus] > 0 ) dest_arr[rem_msg++] = i;
      /* recv_msg_len[i][0] = 0; */
  }
  for (i= Self+1; i < PS ; i++)  {
       if ( send_msg_len[i*PS_plus] > 0 ) dest_arr[rem_msg++] = i;
     /*  recv_msg_len[i][0] = 0;*/
  }
/*  recv_msg_len[Self][0] =0; */
  Datatypesize = datatypesize;
  Recvbuf = recvbuf;
  Recv_msg_len = recv_msg_len;
  next = lrand48() % PS;
  while ( rem_msg > 0 )  {
    next = ( next + 1 ) % rem_msg;
    dest = dest_arr[next];
       /*-- send a request to the current destination --*/
    get_reply = 0;
    Gen_request(dest,Two_Stg_Request,Self,send_msg_len[dest*PS_plus]*datatypesize);
       /*-- Wait for a reply ---*/
    Gen_Poll_until(&get_reply,1);
    if  (busy_dest !=3 ) {   /*  start sending the message */

         /* sending the submessage lengths */
     Gen_send_rport(dest,busy_dest+2,0,&send_msg_len[dest*PS_plus],
                   Destrecv_msg_len,PS+1,sizeof(int));
     ind = 0;
       /* Reshuffling */
     temp = ptrsend;
     for (i=0; i<PS;i++) 
       if (  send_msg_len[dest*PS_plus+(i+1)] > 0 ) {
        memcpy(temp,send_msg_ptr[dest*PS+i],send_msg_len[dest*PS_plus+(i+1)]*datatypesize);
/*      for (k=0;  k < send_msg_len[dest*PS_plus+(i+1)]*datatypesize ; k++)
        temp[ind+k] =  *(send_msg_ptr[dest*PS+i]+k); */
        temp +=  send_msg_len[dest*PS_plus+(i+1)]*datatypesize;
    }
     Gen_send_rport(dest,busy_dest,0,ptrsend,
                         DestRecvBuf,send_msg_len[dest*PS_plus],datatypesize);
     dest_arr[next] = dest_arr[--rem_msg];
     }
   }  /* While loop */
  Set_Global_Or(0);   /* I have done sending all the messages */
  while ( ( busy != 0 )  || ( Get_Global_Or() == 1 ) )
    Gen_Poll();
/* Free rports */
  if ( Free_rport(1)==-1) printf(" Err....");
  if ( Free_rport(2)==-1) printf(" Err2...");
  if ( Free_rport(3)==-1) printf(" Err....");
  if ( Free_rport(4)==-1) printf(" Err2...");
  free((char *) ptrsend);
  free((char *) dest_arr);

  GenBarrier();
}
/*------------  First Stage of the Two stage algorithm  ---------*/

int  first_stage(sendbuf,sendcount,sdispls,datatypesize,Alg,recv_stage1)
void *sendbuf;    /* address of the send buffer */
int  *sendcount;   /* Number of elements send to each processor,
                        This is an array of size PS */
int  *sdispls;   /* the displacement of each send block from the beginning of
               the sendbuf*/
int  datatypesize;   /* the size , in bytes, of the elements */
int  Alg;
char *recv_stage1;
{
  int *send_count_div_P;
  int *send_count_mod_P;
  int i, j, k, q;
  int *more_elements;
  int first_inter_get_excess_elem;
  int next_inter_get_excess_elem;
  int last_inter_get_excess_elem;
  int excess_elements;
  int prefix_sum_excess_elem;
  int temp,PS_plus;
  char *sendptr;

  send_count_div_P = (int *) malloc((unsigned)(sizeof(int)*PS));
  send_count_mod_P = (int *) malloc((unsigned)(sizeof(int)*PS));
  more_elements = (int *) malloc((unsigned)(sizeof(int)*PS));
  excess_elements = 0;
  PS_plus = PS + 1;
  send_count_div_P[Self] = send_count_mod_P[Self] = 0;
  for (i=0; i< Self; i++) {
    send_count_div_P[i] = sendcount[i]/PS;
    send_count_mod_P[i] = sendcount[i]%PS;
    excess_elements += send_count_mod_P[i];
  }
  for (i=Self+1; i< PS; i++) {
    send_count_div_P[i] = sendcount[i]/PS;
    send_count_mod_P[i] = sendcount[i]%PS;
    excess_elements += send_count_mod_P[i];
  }
  /* send_count_mod_P[i] indicates # of excess elements meant for next proc i         */
  /* send_count_div_P[i] is # of elements originally meant for proc i now being sent
     to every intermediate processor                                       */

  /* among the excess elements, others procs need not know abt some        */
  temp = excess_elements / PS;
  for (i=0; i< PS; i++) more_elements[i] = temp;

  excess_elements %= PS;



  prefix_sum_excess_elem = Gen_prefix_sum_int(excess_elements);


  first_inter_get_excess_elem = prefix_sum_excess_elem % PS;
  last_inter_get_excess_elem = first_inter_get_excess_elem +
                               excess_elements -1 ;

  for (j=0; j< PS; j++)  send1_msg_len[j*PS_plus] = 0;
  /* send1_msg_len[j][0] is current message size for intermediate proc j    */
  if ( excess_elements > 0 )
   for (j=0; j<PS; j++)
      if ( ( (j < first_inter_get_excess_elem) &&
             (j <= last_inter_get_excess_elem%PS)) ||   /* with wraparound - top */
        ((j >= first_inter_get_excess_elem)
         && (j <= last_inter_get_excess_elem)))      /* with wraparound-bot, no w */
       more_elements[j]++;

  next_inter_get_excess_elem = first_inter_get_excess_elem;
  for (i=0; i< PS; i++) {       /* i gives destination processor #  */
    for (j=0; j<PS; j++)
      send1_msg_len[j*PS_plus] += send1_msg_len[j*PS_plus+(i+1)] = send_count_div_P[i];
    j = next_inter_get_excess_elem;

    while (send_count_mod_P[i] > 0)      /* send_count_mod_P[i] must be < PS  */
     /* if (more_elements[j] > 0)*/  {
        more_elements[j]--; send_count_mod_P[i]--;
        send1_msg_len[j*PS_plus+(i+1)] += 1;
        send1_msg_len[j*PS_plus] += 1;
        j = (j+1)%PS;
      }
    next_inter_get_excess_elem = j;
    }

  sendptr = sendbuf;
  for (i=0; i< PS; i++) {       /* i - dest proc, j - interm. proc   */
    temp = sdispls[i];
    for (j=0; j< PS; j++) {
      send1_msg_ptr[j*PS+i] = sendptr+(temp*datatypesize);
      temp += send1_msg_len[j*PS_plus+(i+1)];
    }
  }

 free((char *) more_elements); 
 free((char *) send_count_mod_P);

 free((char *) send_count_div_P); 


  switch (Alg)  {
  case 1 :  /*  LP_Xor Without Reshuffling */
          Two_stg_LP_WOR(sendbuf,recv_stage1,send1_msg_len,send1_msg_ptr,
                    recv_msg_len,datatypesize);
          break;
  case 2 :  /*  LP_Xor With Reshuffling */
          Two_stg_LP_WR(sendbuf,recv_stage1,send1_msg_len,send1_msg_ptr,
                    recv_msg_len,datatypesize);
          break;
  case 3:   /* DS without Reshuffling */
          Two_stg_DS_WOR(sendbuf,recv_stage1,send1_msg_len,send1_msg_ptr,
                    recv_msg_len,length,datatypesize);
          break;
  case 4:   /* DS with Reshuffling */
          Two_stg_DS_WR(sendbuf,recv_stage1,send1_msg_len,send1_msg_ptr,
                    recv_msg_len,length,datatypesize);
          break;
  }
}

/*-----------  Second stage of the Two stage algorithm  -------------*/

int  second_stage(recv_stage1,recv_stage2,Alg,datatypesize)
char  *recv_stage1;
char  *recv_stage2;
int   Alg;
int   datatypesize;
{
 int i,j,total,ind;
 int PS_plus;

 PS_plus = PS +1;
 for (i=0; i< PS; i++)   /* initialize the total length of each message */
   send2_msg_len[i*PS_plus] = 0;

  for (i=0; i< PS; i++)  /* j: dest proc, i: source proc */
   if (i != Self) {
    ind=length*i;   /* the start position of the current source message*/
    for (j=0; j< PS; j++) {
      total = recv_msg_len[i*PS_plus+(j+1)]; /* the length of the current submessage */

      send2_msg_ptr[j*PS+i] = (char *) (recv_stage1+ind*datatypesize);
      ind += total;
      send2_msg_len[j*PS_plus] += total;
      send2_msg_len[j*PS_plus+(i+1)] = total;
    }
  } else {
    for (j=0; j< PS; j++) {
      total = send1_msg_len[i*PS_plus+(j+1)]; /* the length of the current submsg */

      send2_msg_ptr[j*PS+i] = send1_msg_ptr[i*PS+j];
      send2_msg_len[j*PS_plus] += total;
      send2_msg_len[j*PS_plus+(i+1)] = total;
    }
  }
  switch (Alg)  {
  case 1 :  /*  LP_Xor Without Reshuffling*/
          Two_stg_LP_WOR(recv_stage1,recv_stage2,send2_msg_len,send2_msg_ptr,
                    recv_msg_len,datatypesize);
          break;
  case 2 :  /*  LP_Xor With Reshuffling*/
          Two_stg_LP_WR(recv_stage1,recv_stage2,send2_msg_len,send2_msg_ptr,
                    recv_msg_len,datatypesize);
          break;
  case 3:   /* DS without Reshuffling */
          Two_stg_DS_WOR(recv_stage1,recv_stage2,send2_msg_len,send2_msg_ptr,
                    recv_msg_len,length,datatypesize);
          break;
  case 4:   /* DS with Reshuffling */
          Two_stg_DS_WR(recv_stage1,recv_stage2,send2_msg_len,send2_msg_ptr,
                    recv_msg_len,length,datatypesize);
          break;
  }
}

/*-------------   Reshuffling of the Two stage algorithm  -----------*/

int  reshuffling(recv_stage2,recvbuf,rdispls,datatypesize)
char  *recv_stage2;
void  *recvbuf;
int   *rdispls;
int datatypesize;
{
 char *ptrrecv;
 int  i,j,k;
 int total,ind,PS_plus;
 int *ptr;
int sum=0;

  PS_plus = PS + 1;
  ptr = (int *) malloc((unsigned)(sizeof(int)*PS));
  for (i=0 ; i < PS ; i++)
    ptr[i] = length*i;
  for (i=0 ; i < PS ; i++)  { /* source processor */
   ind = 0;
   if ( i != Self )  {
   for (j=0 ; j < PS ; j++)  { /* intermediate processor */

       if  (j!= Self) {
         if ( recv_msg_len[j*PS_plus+(i+1)] > 0 )  {
         total = recv_msg_len[j*PS_plus+(i+1)];

sum+=total;
        ptrrecv = recvbuf;
        ptrrecv += (rdispls[i]+ind)*datatypesize;
        memcpy(ptrrecv,&recv_stage2[ptr[j]*datatypesize],total*datatypesize);
/*         for (k=0; k < total*datatypesize ; k++)
           ptrrecv[(rdispls[i]+ind)*datatypesize+k] =
              recv_stage2[ptr[j]*datatypesize+k];  */
         ptr[j] += total;
         ind += total;
	 }
       }
      else 
         if ( send2_msg_len[Self*PS_plus+(i+1)] > 0 ) {
         total = send2_msg_len[Self*PS_plus+(i+1)];
sum+=total;
         ptrrecv = recvbuf;
         ptrrecv += (rdispls[i]+ind)*datatypesize;
         memcpy(ptrrecv,send2_msg_ptr[j*PS+i],total*datatypesize);
/*         for (k=0 ; k < total*datatypesize ; k++)
           ptrrecv[(rdispls[i]+ind)*datatypesize+k] =
             (char) *(send2_msg_ptr[j*PS+i]+k); */
         ind += total;
       }
     }
   }
 }
/*  if ( sum != 31 ) printf(" Er   3..%d \n",Self); */ 
}
            /*-----  Two  Stage Algorithm  -----*/

int Two_Stage_Many_Con(sendbuf,sendcount,sdispls,datatypesize,recvbuf,recvcount,
                     rdispls,Alg1,Alg2)
  void *sendbuf;    /* address of the send buffer */
  int  *sendcount;   /* Number of elements send to each processor,
                        This is an array of size PS */
  int  *sdispls;   /* the displacement of each send block from the beginning of
               the sendbuf*/
  int  datatypesize;   /* the size , in bytes, of the elements */
  void *recvbuf;   /* address of the receive buffer */
  int  *recvcount;   /* No. of elements receive from each processor.
                        This is an array of size PS */
  int  *rdispls;   /* the displacement of each recv block from the beginning
                      of the recvbuf */
  int  Alg1;    /* algorithm for stage 1 */
  int  Alg2;   /* algorithm for stage 2 */
{
  int i,max,temp;

    /* Find the Traffic bound */
  max = sendcount[0];
  temp  = recvcount[0];
  for (i=1; i < PS ; i++) {
    max += sendcount[i];
    temp += recvcount[i];
  }
  if ( temp > max ) max = temp;
  max = Gen_reduce_max_int(max);  /* The bound */

 /*??? Make  Sure ......*/
  max += (PS*PS+PS);  /* to handle the excess elements */

  recv_stage1 = (char *) malloc((unsigned)(datatypesize*max));
  recv_stage2 = (char *) malloc((unsigned)(datatypesize*max)); 
  recv_msg_len = (int *) malloc((unsigned)(sizeof(int)*PS*(PS+1)));
  send1_msg_len = (int *) malloc((unsigned)(sizeof(int)*PS*(PS+1)));
  send2_msg_len = (int *) malloc((unsigned)(sizeof(int)*PS*(PS+1)));
  send1_msg_ptr = (charptr *) malloc((unsigned)(sizeof(charptr)*PS*PS));
  send2_msg_ptr = (charptr *) malloc((unsigned)(sizeof(charptr)*PS*PS));
  if ((recv_stage1==NULL)||(recv_stage2==NULL)||(send1_msg_len==NULL)||(send2_msg_len==NULL)||
    (send1_msg_ptr==NULL)||(send2_msg_ptr==NULL)||(recv_msg_len==NULL))
   printf(" MEMEORY SHORTAGE....!!!"); 

  length = max / PS;


  first_stage(sendbuf,sendcount,sdispls,datatypesize,Alg1,recv_stage1);

  second_stage(recv_stage1,recv_stage2,Alg2,datatypesize);


  reshuffling(recv_stage2,recvbuf,rdispls,datatypesize);


  free((char *) send2_msg_ptr);  
  free((char *) send1_msg_ptr);  
  free((char *) send2_msg_len);
  free((char *) send1_msg_len);
  free((char *) recv_msg_len);
  free((char *) recv_stage2); 
  free((char *) recv_stage1);






  GenBarrier();
}


/**************************************************************************
                Function:  raw
                Purpose :  perfrom rarw where the values are ones,
                           using the two-stage algorithm.
                Author  :  Khaled Alsabti
                Date    :  12/30/95
**************************************************************************/

/******************************************************************/
/*        Random Access Write Using Two Stage Algorithm           */
/******************************************************************/
int   RAW_2Stg(
int indicies[],              /* the addresses */
int NP,                     /* the number of elements */             
int MP,                     /* the local size of the histogram */
int hist[],                 /* the histogram (output )*/
int algorithm)
{
  int K;   /*  K is the number of buckets at each processor  */
  int *count_bucket;  /* holds the number of values to each bucket */
  int *sum_count_bucket;
  int *prefix_sum_count_bucket;
  int *prefix_count_bucket,*bucket_Start;
  int *ptr_bucket,*temp_ptr_bucket;
  
  int *sendcount;
  int *sdispls;
  int *recvcount;
  int *rdispls;
  int *recvbuf,*sendbuf;
  int *dest_ptr;
  
  /*  For Sparse Buckets  */
  int total_sendSP,total_recvSP;
  int *sendcountSP;
  int *sdisplsSP;
  int *recvcountSP;
  int *rdisplsSP;
  int *recvbufSP,*sendbufSP;
  int *dest_ptrSP;
  int *Sparse;
  
  int  KPS;  /* K*PS */
  double S;
  int  NP_div_K;  /* NP/K */
  int  MP_div_K,shift_amount;
  int  Mask_MP_div_K,Mask_MP;

  int total_send,total_recv;
  /*-- Second stage variables --*/
  int start,end,no_buckets,j,i;
  int first_bucket,last_bucket;
  int first_bucket_partial,last_bucket_partial;
  int B,inter_dest,ind,first_full_bucket,last_full_bucket,Size;
  int first_proc,last_proc,no_proc,dest,source,end_bucket_s,end_s;
  int *proc_list,*temp;
  int *sendbuf2,*recvbuf2;
  int ind1,ind2,first,last,last_plus1,logNP;
  /*-- determining the number of buckets  --*/
  
  
  S =1;
  K = 4;
  KPS = K*PS;
  NP_div_K = NP/K * S;;
  MP_div_K = MP/K;
  Mask_MP_div_K = MP_div_K-1;
  Mask_MP = MP-1;
#if Machine == Paragon
  shift_amount = log(MP*PS)/log(2) - log(KPS)/log(2);
  logNP = log(NP)/log(2);
#elif MACHINE == nCube
  shift_amount = log(MP*PS)/log(2) - log(KPS)/log(2);
  logNP = log(NP)/log(2);
#else
  shift_amount = log2((double)MP*PS)-log2((double)KPS);
  logNP = log2((double)NP);
#endif
  dest_ptr = (int *) malloc(PS*sizeof(int));
  rdispls = (int *) malloc(PS*sizeof(int));
  recvcount = (int *) malloc(PS*sizeof(int));
  sdispls = (int *) malloc(PS*sizeof(int));
  sendcount = (int *) malloc(PS*sizeof(int));
  ptr_bucket = (int *) malloc(KPS*sizeof(int));
  prefix_sum_count_bucket = (int *) malloc(KPS*sizeof(int));
  sum_count_bucket = (int *) malloc(KPS*sizeof(int));
  prefix_count_bucket = (int *) malloc(KPS*sizeof(int));
  count_bucket = (int *) malloc(KPS*sizeof(int));
  temp_ptr_bucket = (int *) malloc(KPS*sizeof(int));
  dest_ptrSP = (int *) malloc(PS*sizeof(int));
  rdisplsSP = (int *) malloc(PS*sizeof(int));
  recvcountSP = (int *) malloc(PS*sizeof(int));
  sdisplsSP = (int *) malloc(PS*sizeof(int));
  sendcountSP = (int *) malloc(PS*sizeof(int));
  Sparse = (int *) malloc(KPS*sizeof(int));
  bucket_Start = (int *) malloc(KPS*sizeof(int));
if ( bucket_Start == NULL ) { printf(" Out of Memory... in RAW\n"); exit(0); }		
     /****************       First  Stage          *************/
      /*---------     Preprocessing of First  Stage   --------*/
  /*-- Counting the number of threads writing to each bucket ---*/
  for (i=0; i < KPS ; i++)
       count_bucket[i] = 0;
  for (i=0; i < NP ; i++)
   hist[i] = 0; /* INITIAL VALUE */
  for (i=0 ; i < NP ; i++)
     count_bucket[indicies[i]>>shift_amount]++;

  /*-- Performing a global vector sum-combine and prefix-sumon count_bucket --*/
  Gen_global_reduce_prefix_sum_v(prefix_count_bucket,sum_count_bucket,
                        count_bucket,KPS);
  /*-- Performing a local execlusive prefix-sum on sum_count_bucket --*/
  prefix_sum_count_bucket[0] = 0;
  for (i=0 ; i < KPS-1 ; i++)  {
    prefix_sum_count_bucket[i+1] = sum_count_bucket[i]+
                                prefix_sum_count_bucket[i];
    ptr_bucket[i] = prefix_count_bucket[i] + prefix_sum_count_bucket[i];
  }
  ptr_bucket[i] = prefix_count_bucket[i] + prefix_sum_count_bucket[i];

  for (i=0; i < PS ; i++)
     sendcountSP[i] = sendcount[i] = 0;
  ind1 = ind2 = 0;
  for (i=0 ; i < KPS ; i++)
   if ( sum_count_bucket[i] <= NP_div_K )  {
     Sparse[i] = 1;
     dest = i/K;
     sendcountSP[dest] += count_bucket[i];
     sum_count_bucket[i] = 0;
     bucket_Start[i] = ind1;
     ind1+= count_bucket[i];
   }
   else {
     Sparse[i] = 0;
     bucket_Start[i] = ind2;
     ind2 += count_bucket[i];
     first = ptr_bucket[i] >> logNP;
     last = (ptr_bucket[i]+count_bucket[i]-1) >> logNP;
     if ( first == last ) sendcount[first]+= count_bucket[i];
     else {
       sendcount[first] += NP- ptr_bucket[i]%NP;
       sendcount[last] += (ptr_bucket[i]+count_bucket[i])%NP;
     }
   } 
		
   /*--  Combine the receive vector, recvcount, --*/
  preprocessing_comm(sendcount,recvcount,sdispls,rdispls,&total_send,&total_recv);

  preprocessing_comm(sendcountSP,recvcountSP,sdisplsSP,rdisplsSP,&total_sendSP,&total_recvSP);

    /* we send only the  index */
  sendbuf = (int *) malloc(total_send*sizeof(int));
  recvbuf = (int *) malloc(total_recv*sizeof(int));

  sendbufSP = (int *) malloc(total_sendSP*sizeof(int));
  recvbufSP = (int *) malloc(total_recvSP*sizeof(int));

       /*-- Building the send buffer  --*/

  for (i=0 ; i < NP ; i++)  {
   B = indicies[i]>>shift_amount;  /* this value is for bucket B */
   if ( Sparse[B] == 0 )  /* dense bucket */
    sendbuf[bucket_Start[B]++] = indicies[i];
   else
    sendbufSP[bucket_Start[B]++] = indicies[i]&Mask_MP;
 }
      /*   For Sparse Buckets  */
   ManytoMany_Con(sendbufSP,sendcountSP,sdisplsSP,sizeof(int),
                  recvbufSP,recvcountSP,rdisplsSP,algorithm,0,0);
       /*--   First stage communcations  --*/
   ManytoMany_Con(sendbuf,sendcount,sdispls,sizeof(int),
                  recvbuf,recvcount,rdispls,algorithm,0,0);



   for (i=0 ; i < total_recvSP ; i++)
      hist[recvbufSP[i]] ++;

     /****************      Second  Stage          *************/
        /*-- Determining The number of partial buckets,2,1 or 0 and the
             total number of buckets   --*/
   no_buckets =0;
   start = Self*NP;
   end = (Self+1)*NP-1;
   ind = KPS-1;

   while ( (ind> 0 ) &&  (prefix_sum_count_bucket[ind] > start ))
     ind--;
   first_bucket = ind;
   while ( (ind <KPS) && (prefix_sum_count_bucket[ind] <= end )) {
     no_buckets++;
     ind++;
    }
   last_bucket = ind -1;

     /*-- determine if the first bucket is partial --*/
   if ( (prefix_sum_count_bucket[first_bucket] < start ) &&
        ( Sparse[first_bucket] == 0) &&
/* ??*/ (prefix_sum_count_bucket[first_bucket]+sum_count_bucket[first_bucket] > start))
     first_bucket_partial = 1;  /* Yes */
   else
     first_bucket_partial = 0;  /* No */

     /*-- determine if the last bucket is partial --*/
   if (((prefix_sum_count_bucket[last_bucket]+sum_count_bucket[last_bucket]-1)
      > end ) && (Sparse[last_bucket] == 0))
     last_bucket_partial = 1; /* Yes */
   else
     last_bucket_partial = 0; /* NO */
     /*-- Calculating the size of the sending buffer */
   if ( first_bucket_partial )  first_full_bucket = first_bucket + 1;
   else first_full_bucket = first_bucket;
   if ( last_bucket_partial ) last_full_bucket = last_bucket -1;
   else last_full_bucket = last_bucket;
   if ( (last_bucket_partial)||(first_bucket_partial))
     Size = MP_div_K;
   else Size = 0;
   if ( (last_bucket_partial)&&(first_bucket_partial)&& ( first_bucket != last_bucket ))
     Size += MP_div_K;
   ind = 0;

   for (i=0; i < PS ; i++) {
     sendcount[i] = 0;
     sdispls[i] = -1;
     }

   for (i=first_full_bucket;  i <= last_full_bucket ; i++)
     if ( Sparse[i] == 0 ) {  /* dense bucket */
       ptr_bucket[i] = ind;
       dest = i/K;
       Size += MP_div_K;
       sendcount[dest] += MP_div_K;
       if ( sdispls[dest] == -1 ) sdispls[dest] = ind;
       ind += MP_div_K;
      }

   if ( (last_bucket_partial )&&((first_bucket_partial==0)||(
        last_bucket!=first_bucket)))   {
     ptr_bucket[last_bucket] = ind;
     dest = last_bucket/K;
     sendcount[dest] += MP_div_K;
     if ( sdispls[dest] == -1 )  sdispls[dest]= ind;
     ind += MP_div_K;
   }
   if  (first_bucket_partial)
      ptr_bucket[first_bucket] = ind;


   sendbuf2 = (int *) malloc(Size*sizeof(int));

   /*-- Initialize sendbuf2 buckets for full bucket and last_bucket if it is
        partial --*/
   for (i=0; i < Size ; i++) sendbuf2[i] = 0;

   /*-- copy the values to the sending buffer and resolve collision --*/
   for (i=0 ; i < total_recv ; i++)  {
     B = recvbuf[i] >> shift_amount;
     ind = recvbuf[i] & Mask_MP_div_K;
     sendbuf2[ptr_bucket[B]+ind] ++;
   }

     /*-- Perform segmented scan  , if any --*/
  if ( (first_bucket_partial) &&
      ( (last_bucket_partial==0) || (first_bucket == last_bucket)) )
    B = first_bucket;
  else
    if ( last_bucket_partial )  B = last_bucket;
    else B = -1;
  source = -1;
  if ( B != -1 ) {  /*  I am invloved in a segemented scan */
    
    /*-- determine the processors invloving in the segemented scan --*/
    first_proc = prefix_sum_count_bucket[B] / NP;
    last_proc = (prefix_sum_count_bucket[B]+sum_count_bucket[B]-1)/NP;
    no_proc = last_proc - first_proc + 1;
    
    /* Check if last_proc has succeeding partial bucket, if so then
       delete it from the list */
    /* determine if last proc  has succeeding partial bucket */
    end_s = ( last_proc + 1 ) * NP - 1;
    ind = B;
    while ( (ind < KPS ) && ( prefix_sum_count_bucket[ind] <= end_s) )
      ind++;
    end_bucket_s = ind -1;
    if ( ( (prefix_sum_count_bucket[end_bucket_s] +
            sum_count_bucket[end_bucket_s] -1 ) > end_s ) &&
	( Sparse[end_bucket_s] == 0 ) &&
	( B != end_bucket_s ) )  {
      no_proc--;  
      if ( Self == first_proc ) source = last_proc;
      last_proc--;
    }
    
    proc_list = (int *) malloc(no_proc*sizeof(int));
    for (i=0 ; i < no_proc ; i++)
      proc_list[i] = first_proc + i;

    if ( no_proc > 1 )

      /* change segemnted_sum_v to deal with only 1 dim. */
      segmented_sum_v(&sendbuf2[ptr_bucket[B]],MP_div_K,proc_list,no_proc);
  }
  GenBarrier();


  /*-- send/receive the preceding bucket , if any --*/
  /*-- sending the preceeding partial bucket, if needed --*/
  if ( (first_bucket_partial==1) && (last_bucket_partial==1) &&
      ( first_bucket != last_bucket ))  {
    dest = prefix_sum_count_bucket[first_bucket]/NP;
    GenSynSend(&sendbuf2[ptr_bucket[first_bucket]],MP_div_K*sizeof(int),dest,Self);
  }

  /*-- receiving a bucket , if any --*/
  if ( source != -1 ) {

    temp = (int *) malloc(MP_div_K*sizeof(int));
    GenSynReceive(temp,MP_div_K*sizeof(int),source,source);
    for (i=0 ; i < MP_div_K ; i++)
      sendbuf2[ptr_bucket[last_bucket]+i] += temp[i];
  }
  GenBarrier();
  /*--     Second Stage  Communication --*/
  /*  Combine the receive vector, recvcount, */

  AlltoAll(sendcount,1,sizeof(int),recvcount,1);  /*  Using AC */

  /*-- Calculating  rdispls   --*/
  rdispls[0] = 0;
  total_recv = recvcount[0];
  for (i=1 ; i < PS ; i++)  {
    rdispls[i] = rdispls[i-1] + recvcount[i-1];
    total_recv += recvcount[i];
  }

  recvbuf2 = (int *) malloc(total_recv*sizeof(int));
  ManytoMany_Con(sendbuf2,sendcount,sdispls,sizeof(int),
		 recvbuf2,recvcount,rdispls,algorithm,0,0);
  
  /*--   Copying the values to the final destination  --*/
  first = Self*K;
  last_plus1 = first+K;
  j = 0;
  for ( i=first ;  i < last_plus1; i++)
    if ( Sparse[i] == 0 )  { /* dense bucket */
      ind = MP_div_K * (i%K );
      end = ind+MP_div_K;
      for ( ; ind < end ; ind++,j++)
	hist[ind] += recvbuf2[j];
    }

  /*-- House Keeping --*/
  free(dest_ptr); 
  free(rdispls); 
  free(recvcount); 
  free(sdispls); 
  free(sendcount);
  free(ptr_bucket); 
  free(prefix_sum_count_bucket); 
  free(sum_count_bucket); 
  free(prefix_count_bucket); 
  free(count_bucket); 
  free(temp_ptr_bucket); 
  
  free(dest_ptrSP); 
  free(rdisplsSP); 
  free(recvcountSP);
  free(sdisplsSP); 
  free(sendcountSP);
  free(Sparse);
  free(bucket_Start); 
  GenBarrier();
}

/**************************************************************************
                Function:  raw
                Purpose :  perfrom rarw where the values are ones,
                           using the directalgorithm.
                Author  :  Khaled Alsabti
                Date    :  1/22/96
**************************************************************************/
 
/*------   Random Access Write  Using Direct algorithm   ---------*/
int RAW_Dir(
int  *indicies,  /*   List of N/PS indices, one for each value. These indices 
                 are global */
int  NP,    /*  This is the number of values.*/
int  MP,    /* the size of the histogram */
int  *Hist,  /*  List of N/PS Hist */
int  algorithm)
{
  int i,dest;
  int  *dest_ptr;
  int  *sendcount,*recvcount,*sdispls,*rdispls;
  int  total_send,total_recv;
  int  *sendbuf,*recvbuf;


  sendcount  = (int *) malloc(sizeof(int)*PS);
  recvcount  = (int *) malloc(sizeof(int)*PS);
  sdispls    = (int *) malloc(sizeof(int)*PS);
  rdispls    = (int *) malloc(sizeof(int)*PS);
  dest_ptr   = (int *) malloc(sizeof(int)*PS);    
  for (i=0 ;  i < NP  ; i++)  /* Initializing the Hist */
    Hist[i] = 0;  /* THE INITIAL VALUE */
  for (i=0 ; i < PS ; i++) 
    sendcount[i] = 0; 

  for (i=0 ; i < NP ; i++)  /* Counting the No. of values to be
                               sent to each processor */
    { dest = indicies[i]/MP;
    sendcount[dest]++; }

  /*  Gather the receiving vector */
  preprocessing_comm(sendcount,recvcount,sdispls,rdispls,&total_send,&total_recv);

 /*  dest_ptr is used to fill the sendbuf */
  for (i=0; i < PS ; i++)
   dest_ptr[i] = sdispls[i];

  /* -- Allocations for sendbuf and recvbuf  */
  sendbuf = (int *) malloc(total_send*sizeof(int));
  recvbuf = (int *) malloc(total_recv*sizeof(int));
 if ( ((sendbuf==NULL) &&(total_send>0)) || ( (recvbuf==NULL) &&(total_recv>0))) {
   printf("  Memory Shortage....\n");

 } 
  /* Building the sendbuf  */
 for (i= 0 ; i < NP ; i++) {

  dest= indicies[i]/MP;
  sendbuf[dest_ptr[dest]++] = indicies[i]% MP; /* Now the index is local */
 }

 /*--- Performing Many-to-Many using AC   ----*/

 ManytoMany_Con(sendbuf,sendcount,sdispls,sizeof(int),
   recvbuf,recvcount,rdispls,algorithm,0,0);


 /*---  Update the local Hist and in case of collision, resolve it ----*/

 for (i = 0; i < total_recv ; i++) {
   /*-- Here Plus is the collision resolution strategy----*/
  Hist[recvbuf[i]] ++;
 }   

 /*--- Hosue Keeping  ---*/
/* free((int *) recvbuf);
 free((int *) sendbuf);
 free((int *) dest_ptr);
 free((int *) rdispls);
 free((int *) sdispls);
 free((int *) recvcount);
 free((int *) sendcount); */
 GenBarrier();

}  /*  RAW_AC  */
